#ifndef __XTREE
#define __XTREE

#include "linlist.h"
//////////////////////////////////////////////////////////////////////////////
// defines
//////////////////////////////////////////////////////////////////////////////
#define MAXREAL 9.99e20
#define MAX_DIMENSION 256
#define EXT_SIZE 5
#define MAX_OVERLAP 0.2
#define MIN_FANOUT 0.35

#ifndef S3
#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif  // TRUE
#define min(a, b) (((a) < (b))? (a) : (b)  ) 
#define max(a, b) (((a) > (b))? (a) : (b)  ) 
#endif  // S3


//////////////////////////////////////////////////////////////////////////////
// C-prototypes
//////////////////////////////////////////////////////////////////////////////
float area(int dimension, float *mbr);
float margin(int dimension, float *mbr);
float overlap(int dimension, float *r1, float *r2);
float OBJECT_DIST(const void  *Point1,const void  *Point2);
float MINMAXDIST(const void  *Point, float *bounces);
float MINDIST(const void *Point, float *bounces);

bool inside(float &p, float &lb, float &ub);
void enlarge(int dimension, float **mbr, float *r1, float *r2);
bool is_inside(int dimension, float *p, float *mbr);
bool section(int dimension, float *mbr1, float *mbr2);
bool section_c(int dimension, float *mbr1, const void *center, float radius);

void check_mbr(int dimension, float *mbr);

int sort_lower_mbr(const void *d1, const void *d2);
int sort_upper_mbr(const void *d1, const void *d2);
int sort_center_mbr(const void *d1, const void *d2);
int sort_min_dist(const void *element1, const void *element2);
int prune_branch_list(float *nearest_distanz, const void *activebrunchList, int n);
int test_branch_list(const void *abL, const void *sL, int n, int last);

#ifndef S3
void error(char *t, bool ex);
#endif

//////////////////////////////////////////////////////////////////////////////
// externals
//////////////////////////////////////////////////////////////////////////////
extern int XTDirNode__dimension;
extern void* XTDirNode__my_tree;
extern int XTDataNode__dimension;
extern float page_access;

//////////////////////////////////////////////////////////////////////////////
// enums / types
//////////////////////////////////////////////////////////////////////////////
enum SECTION {OVERLAP, INSIDE, S_NONE};
enum R_OVERFLOW {SPLIT, SUPERNODE, NONE};
typedef float *floatptr;
typedef int bitvector;

//////////////////////////////////////////////////////////////////////////////
// classes
//////////////////////////////////////////////////////////////////////////////
#ifdef S3
class Index;
class Result;
#endif // S3

struct SortMbr
{
    int dimension;
    float *mbr;
    float *center;
    int index;
};

struct DataStruct
{
    int dimension;
    float *Data;
};


struct BranchList
{
    int entry_number;
    float mindist;
    float minmaxdist;
    bool section;
};

template <class DATA> 
class XTNode
{
    friend class XTree<DATA>;
protected:
   XTree<DATA> *my_tree;                // pointer to X-tree
    int capacity;                       // max. # of entries
    int dimension;
    int num_entries;                    // # of used entries
    bool dirty;                         // TRUE, if node has to be written
public:
    int block;
    char level;                         // level of the node in the tree

    int get_num()                       // returns # of used entries
    { return num_entries;}
    virtual int get_num_of_data() = 0;  // returns number of data entries
                                        // behind that node

    virtual DATA * get(int i) = 0;      // returns the i-th object in the 
                                        // tree lying behind that node
    
    virtual bool is_data_node() = 0;	// returns TRUE, if "this" is 
                                        // *XTDataNode 

    virtual float *get_mbr() = 0;       // returns mbr enclosing whole page
    virtual void print()                // prints rectangles
    { }

    int topological_split(float **mbr, int **distribution, int *dim);
                                        // splits an Array of mbr's into 2 
                                        // and returns in distribution
                                        // which *m mbr's are to be
                                        // moved
    int overlap_free_split(float **mbr, int *split_hist, int **distribution, int *dim);

#ifdef S3
    virtual void neighbours(LinList<DATA> *sl,   
		    float eps,           // berechnet fuer alle DATAs in
		    Result *rs,          // sl die Nachbarn, die in der eps-
		    norm_ptr norm) = 0;  // Umgebung liegen
#endif // S3
    virtual void nearest_neighbour_search(DATA *Point, DATA  *Nearest,float *nearest_distanz) = 0;


    virtual void nearest_neighbour_search(DATA *Point, 
		SortedLinList<DATA> *res, float *nearest_distanz) =  0;

    virtual void range_query(float *mbr, SortedLinList<DATA> *res) = 0;

    virtual void range_query(DATA *center, float radius, SortedLinList<DATA> *res) = 0; 

    virtual void overlapping(float *p, int *nodes_t) = 0;
    virtual void nodes(int *nodes_a) = 0;

    virtual void write_info(FILE *f) = 0;
	
    virtual R_OVERFLOW insert(DATA *d, XTNode<DATA> **sn, int *dim) = 0; 
                                        // inserts d recursivly, if there
                                        // occurs a split, FALSE will be
                                        // returned and in sn a 
                                        // pointer to the new node

    virtual void region(float *mbr) = 0;
                                        // prints all entries sectioning
                                        // mbr
    virtual void point_query(float *p) = 0; 
                                        // prints all entries equal to p
                                        
 
    XTNode(XTree<DATA> *rt);
    XTNode(XTree<DATA> *rt, int _block);
    virtual ~XTNode();
};

template <class DATA> 
class XTDirNode: public XTNode<DATA>
{
    friend class XTree<DATA>;

    int nodesize;                       // size of Directory Node
                                        // nodesize > 1 => SuperNode

    DirEntry<DATA> *entries;            // array of entries
    bool son_is_data;                   // TRUE, if son is a data page
public:
    int* intern_block;                  // array of blocks
    int block_capacity;

    void read_header(char* buffer);     // reads header of buffer
    void read_from_buffer(char *buffer, int offset);// reads data from buffer
    void write_to_buffer(char *buffer, int offset); // writes data to buffer

    bool is_data_node()
    {return FALSE;}

    int get_num_of_data();              // returns number of data entries
                                        // behind that node

    DATA * get(int i);                  // returns the i-th object in the 
                                        // tree lying behind that node

    void print();                       // prints rectangles

    float *get_mbr();                   // returns mbr enclosing whole page

    float *calc_mbr(DirEntry<DATA> *de, int num_entries);
    					// calculates mbr enclosing DirEntries
    
    void enter(DirEntry<DATA> *de);     // inserts new entry

    void adapt_node(int mult);                 // adapt node to supernode
  
    bool split(XTDirNode<DATA> **sn, int *dim);
    					// splits directory page to *this and sn
    					// returns FALSE, if supernode was created
    int choose_subtree(float *brm);     // chooses best subtree for insertion
#ifdef S3
    void neighbours(LinList<DATA> *sl,  // berechnet fuer alle DATAs in
		    float eps,          // sl die Nachbarn, die in der eps-
		    Result *rs,         // Umgebung liegen
		    norm_ptr norm);
#endif // S3

    void nearest_neighbour_search(DATA *QueryPoint, 
				DATA *Nearest,float *nearest_distanz);

    void nearest_neighbour_search(DATA *QueryPoint, 
			SortedLinList<DATA> *res,
			float *nearest_distanz);

    void range_query(float *mbr, SortedLinList<DATA> *res);

    void range_query(DATA *center, float radius, SortedLinList<DATA> *res);	

    R_OVERFLOW insert(DATA *d, XTNode<DATA> **sn, int *dim);  
                                        // inserts d recursivly, if there
                                        // occurs a split, FALSE will be
                                        // returned and in sn a 
                                        // pointer to the new node
    void region(float *mbr);            // prints all entries sectioning mbr
    void point_query(float *p);         // prints all entries equal to p

    void overlapping(float *p, int *nodes_t);
    void nodes(int *nodes_a);
    void write_info(FILE  *f);

    XTDirNode(int _nodesize, XTree<DATA> *rt);                  
    XTDirNode(XTree<DATA> *rt, int _block);
    virtual ~XTDirNode();
};

template <class DATA> 
class XTDataNode: public XTNode<DATA>
{
    DATA *data;                         // array of data
public:
    void read_from_buffer(char *buffer);// reads data from buffer
    void write_to_buffer(char *buffer); // writes data to buffer

    bool is_data_node()
    {return TRUE;}

    void print();                       // prints vector

    int get_num_of_data();              // returns number of data entries
                                        // in that node
    DATA * get(int i);                  // returns the i-th object in this node 

    float *get_mbr();                   // returns mbr enclosing whole page 

    void split(XTDataNode<DATA> *sn, int *dim);
    					// splits data page into sn and *this;
                                        // returns split dimension
#ifdef S3
    void neighbours(LinList<DATA> *sl,  // berechnet fuer alle DATAs in
		    float eps,          // sl die Nachbarn, die in der eps-
		    Result *rs,         // Umgebung liegen
		    norm_ptr norm);
#endif // S3
    void nearest_neighbour_search(DATA *QueryPoint, 
				DATA *Nearest, float *nearest_distanz);

    void nearest_neighbour_search(DATA *QueryPoint, 
			SortedLinList<DATA> *res, float *nearest_distanz);

    void range_query(float *mbr, SortedLinList<DATA> *res);

    void range_query(DATA *center, float radius, SortedLinList<DATA> *res);

    R_OVERFLOW insert(DATA *d, XTNode<DATA> **sn, int *dim);  
                                        // inserts d recursivly, if there
                                        // occurs a split, FALSE will be
                                        // returned and in sn a 
                                        // pointer to the new node

    void region(float *mbr);            // prints all entries sectioning
                                        // mbr
    void point_query(float *mbr);       // prints all entries equal to p

    void overlapping(float *p, int *nodes_t)
    { nodes_t[0]++; }

    void nodes(int *nodes_a)
    { nodes_a[0]++; }

    void write_info(FILE *f);
    XTDataNode(XTree<DATA> *rt);                  
    XTDataNode(XTree<DATA> *rt, int _block);
    virtual ~XTDataNode();
};


template <class DATA> 
class DirEntry
{
    friend class XTDirNode<DATA>;
    friend class XTree<DATA>;

    XTree<DATA> *my_tree;               // pointer to my Xtree
    int son;                            // block # of son
    XTNode<DATA> *son_ptr;              // pointer to son if in main mem.
    bool son_is_data;                   // TRUE, if son is a data page
    int dimension;                      // dimension of the box
    float *bounces;                     // pointer to box
    int num_of_data;                    // amount of data entries behind the
                                        // son of this entry
    bitvector history;			// split-history of son
public:
    bool is_inside(float *v);           // tests, if v is inside the box
    SECTION section(float *mbr);        // tests, if mbr intersects the box
    XTNode<DATA>* get_son();            // returns pointer to the son,
                                        // loads son if necessary

    bool section_circle(DATA *center, float radius);

    void read_from_buffer(char *buffer);// reads data from buffer   
    void write_to_buffer(char *buffer); // writes data to buffer  

    int get_size();                     // returns amount of needed buffer space

    virtual DirEntry<DATA> & operator = (DirEntry<DATA> &_d);

    DirEntry(int dimension = XTDirNode__dimension, 
	     XTree<DATA> *rt = (XTree<DATA> *)XTDirNode__my_tree);
    virtual ~DirEntry();
};

template <class DATA> 
class XTree 
{
public:
    friend class XTNode<DATA>;
    friend class XTDirNode<DATA>;
    friend class XTDataNode<DATA>;

    int root;                            // block # of root node
    XTNode<DATA> *root_ptr;              // root-node
    bool root_is_data;                   // TRUE, if root is a data page
    int dimension;                       // dimension of the datas

    int num_of_data;	                 // # of stored data 
    int num_of_dnodes;	                 // # of stored data pages
    int num_of_inodes;	                 // # of stored directory pages

    CachedBlockFile *file;	         // storage manager for harddisc blocks

    void load_root();                    // loads root_node into memory
    int akt;                             // # of actually got data (get_next)
    char *header;
    float node_weight[20];               // weight for simulation of cache
    char *block_buffer;                  // buffer for reading and writing nodes 
protected:
    char *user_header;
    void read_header(char *buffer);      // reads XTree header
    void write_header(char *buffer);     // writes XTree header
public:
    int get_num()                        // returns # of stored data
    { return num_of_data; }              

    void insert(DATA *d);                // inserts new data into tree

#ifdef S3
    bool erase(PolygonId &i)             // erases object # i
    { error("XTree::erase: not implemented", FALSE); }
#endif
    DATA *get_first()                    // trace
    { akt = 0; return get(akt); }        // through
    DATA *get_next()                     // all 
    { akt++; return get(akt); }          // data

    DATA *get(int i);                    // get i-th data

    void set_node_weight(int cache_size); // calculate influence of cache
#ifdef S3
    void neighbours(LinList<DATA> *sl,   // berechnet fuer alle DATAs in
		    float eps,           // sl die Nachbarn, die in der eps-
		    Result *rs,          // Umgebung liegen
		    norm_ptr norm);
#endif // S3

    void region(float *mbr);             // print all entries intersection 
                                         // the box mbr
    void point_query(float *p);          // print all entries equal to p

    void nearest_neighbour_search(DATA *QueryPoint, DATA *Nearest);
    
    void nearest_neighbour_search(DATA *QueryPoint, SortedLinList<DATA> *res);

    void k_nearest_neighbour_search(DATA *QueryPoint, int k, SortedLinList<DATA> *NeighborList);

    void range_query(float *mbr, SortedLinList<DATA> *res);

    void range_query(DATA *center, float radius, SortedLinList<DATA> *res);

    void print_info(FILE *f);

    void overlapping(float *p, int *nodes_t);
    void nodes(int *nodes_a);

    void write_info(FILE *f);

    XTree(char *fname, 
	  int _b_length, int cache_size, 
	  int _dimension);
    XTree(char *fname, int cache_size);
    virtual ~XTree();
};


//////////////////////////////////////////////////////////////////////////////
// implementation
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// DirEntry
//////////////////////////////////////////////////////////////////////////////

template <class DATA> 
DirEntry<DATA>::DirEntry(int _dimension, XTree<DATA> *rt)
{
    dimension = _dimension;
    my_tree = rt;
    bounces = new float[2*dimension];
    son_ptr = NULL;
    num_of_data = 0;
    history = 0;    
}

template <class DATA> 
DirEntry<DATA>::~DirEntry()
{
    if (bounces != NULL)
        delete [] bounces;
    if (son_ptr != NULL)
	delete son_ptr;
}

template <class DATA> 
DirEntry<DATA>& DirEntry<DATA>::operator = (DirEntry &_d)
{
    dimension = _d.dimension;
    son = _d.son;
    son_ptr = _d.son_ptr;
    son_is_data = _d.son_is_data;
    num_of_data = _d.num_of_data;
    memcpy(bounces, _d.bounces, sizeof(float) * 2 * dimension);
    history = _d.history;
    
    return *this;
}

template <class DATA> 
bool DirEntry<DATA>::is_inside(float *v)
{
    int i;

    for (i = 0; i < dimension; i++)
    {
	if (v[i] < bounces[2*i] ||     // untere Grenze
	    v[i] > bounces[2*i + 1])   // obere Grenze
	    return FALSE;
    }

    return TRUE;
}

template <class DATA> 
SECTION DirEntry<DATA>::section(float *mbr)
// testet, ob mbr den Eintrag irgendwo ueberlappt
// Uberlappung kann nur dann nicht auf, wenn 
//
//    ttttt eeeee
//
//      oder
//
//    eeeee ttttt
//
// diese beiden Faelle werden fuer jede Dimension getestet
{
    bool inside;
    bool overlap;
    int i;

    overlap = TRUE;
    inside = TRUE;
    for (i = 0; i < dimension; i++)
    {
	if (mbr[2*i]     > bounces[2*i + 1] ||
	    mbr[2*i + 1] < bounces[2*i]) 
	    overlap = FALSE;
	if (mbr[2*i]     < bounces[2*i] ||
	    mbr[2*i + 1] > bounces[2*i + 1]) 
	    inside = FALSE;
    }
    if (inside)
	return INSIDE;
    else if (overlap)
	return OVERLAP;
    else
	return S_NONE;
}


template <class DATA> 
void DirEntry<DATA>::read_from_buffer(char *buffer)
{
    int i;

    i = 2*dimension*sizeof(float);
    memcpy(bounces, buffer, i);
    memcpy(&son, &buffer[i], sizeof(int));
    i += sizeof(int);
    memcpy(&num_of_data, &buffer[i], sizeof(int));
    i += sizeof(int);
    memcpy(&history, &buffer[i], sizeof(int));
}

template <class DATA> 
void DirEntry<DATA>::write_to_buffer(char *buffer)
{
    int i;

    i = 2*dimension*sizeof(float);
    memcpy(buffer, bounces, i);
    memcpy(&buffer[i], &son, sizeof(int));
    i += sizeof(int);
    memcpy(&buffer[i], &num_of_data, sizeof(int));
    i += sizeof(int);
    memcpy(&buffer[i], &history, sizeof(int));
}

template <class DATA> 
int DirEntry<DATA>::get_size()
{
    return 2*dimension * sizeof(float) + sizeof(int) + sizeof(int) + sizeof(int);
}

template <class DATA> 
XTNode<DATA>* DirEntry<DATA>::get_son()
{


    if (son_ptr == NULL)
    {
	if (son_is_data)
	    son_ptr = new XTDataNode<DATA>(my_tree, son);
	else
	    son_ptr = new XTDirNode<DATA>(my_tree, son);
    }
    return son_ptr;
}

template <class DATA>
bool DirEntry<DATA>::section_circle(DATA *center, float radius)
{
 float r2;

 r2 = radius * radius;

 if ( (r2 - MINDIST(center,bounces )) < 1.0e-8)
	return TRUE;
 else
	return FALSE;
}
 

//////////////////////////////////////////////////////////////////////////////
// XTNode
//////////////////////////////////////////////////////////////////////////////

template <class DATA> 
XTNode<DATA>::~XTNode()
{
}

template <class DATA> 
XTNode<DATA>::XTNode(XTree<DATA> *rt)
{
    my_tree = rt; 
    dimension = rt->dimension; 
    num_entries = 0; 

    block = -1;
}


template <class DATA> 
int XTNode<DATA>::topological_split(float **mbr, int **distribution, int *dim)
// teilt ein Array von mbrs in zwei Haelften und liefert in distribution
// zurueck, welche *m Eintraege in die neue Seite verschoben werden muessen
// hierfuer wird der R*-Baum Split Algorithmus benutzt
{
#ifdef SHOWMBR
	split_000++;
#endif

    bool lu;
    int i, j, k, l, s, n, m1, dist, split_axis;
    SortMbr *sml, *smu;
    float minmarg, marg, minover, mindead, dead, over, 
	*rxmbr, *rymbr;

    // how many nodes are used?
    n = get_num();

    // nodes have to be filled at least 40%
    m1 = (int) ((float)n * 0.40); 

    // sort by lower value of their rectangles
    // Indexarray aufbauen und initialisieren
    sml = new SortMbr[n];    
    smu = new SortMbr[n];    
    rxmbr = new float[2*dimension];
    rymbr = new float[2*dimension];

    // choose split axis
    minmarg = MAXREAL;
    for (i = 0; i < dimension; i++)
    // for each axis
    {
        for (j = 0; j < n; j++)
        {
            sml[j].index = smu[j].index = j;
            sml[j].dimension = smu[j].dimension = i;
            sml[j].mbr = smu[j].mbr = mbr[j];
        }

        // Sort by lower and upper value perpendicular axis_i
      	qsort(sml, n, sizeof(SortMbr), sort_lower_mbr);
        qsort(smu, n, sizeof(SortMbr), sort_upper_mbr);

        marg = 0.0;
        // for all possible distributions of sml
        for (k = 0; k < n - 2*m1 + 1; k++)
        {
            // now calculate margin of R1
	    // initialize mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =    MAXREAL;
		rxmbr[s+1] = -MAXREAL;
	    }
            for (l = 0; l < m1+k; l++)
            {
                // calculate mbr of R1 
		for (s = 0; s < 2*dimension; s += 2)
	        {
	            rxmbr[s] =   min(rxmbr[s],   sml[l].mbr[s]);
	            rxmbr[s+1] = max(rxmbr[s+1], sml[l].mbr[s+1]);
                }
            }
	    marg += margin(dimension, rxmbr); 

            // now calculate margin of R2
	    // initialize mbr of R2 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =    MAXREAL;
		rxmbr[s+1] = -MAXREAL;
	    }
            for ( ; l < n; l++)
            {
                // calculate mbr of R1 
		for (s = 0; s < 2*dimension; s += 2)
	        {
	            rxmbr[s] =   min(rxmbr[s],   sml[l].mbr[s]);
	            rxmbr[s+1] = max(rxmbr[s+1], sml[l].mbr[s+1]);
                }
            }
	    marg += margin(dimension, rxmbr); 
        }

        // for all possible distributions of smu
       	for (k = 0; k < n - 2*m1 + 1; k++)
        {
            // now calculate margin of R1
	    // initialize mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =    MAXREAL;
		rxmbr[s+1] = -MAXREAL;
	    }
            for (l = 0; l < m1+k; l++)
            {
                // calculate mbr of R1 
		for (s = 0; s < 2*dimension; s += 2)
	        {
	            rxmbr[s] =   min(rxmbr[s],   smu[l].mbr[s]);
	            rxmbr[s+1] = max(rxmbr[s+1], smu[l].mbr[s+1]);
                }
            }
	    marg += margin(dimension, rxmbr); 

            // now calculate margin of R2
	    // initialize mbr of R2 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =    MAXREAL;
		rxmbr[s+1] = -MAXREAL;
	    }
            for ( ; l < n; l++)
            {
                // calculate mbr of R1 
		for (s = 0; s < 2*dimension; s += 2)
	        {
	            rxmbr[s] =   min(rxmbr[s],   smu[l].mbr[s]);
	            rxmbr[s+1] = max(rxmbr[s+1], smu[l].mbr[s+1]);
                }
            }
	    marg += margin(dimension, rxmbr); 
        }

        // actual margin better than optimum?
        if (marg < minmarg)
        {
            split_axis = i;
            minmarg = marg;
        }

    }

    /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
    *dim = split_axis;

    
    // choose best distribution for split axis
    for (j = 0; j < n; j++)
    {
	sml[j].index = smu[j].index = j;
	sml[j].dimension = smu[j].dimension = split_axis;
	sml[j].mbr = smu[j].mbr = mbr[j];
    }
	       
    // Sort by lower and upper value perpendicular split axis
    qsort(sml, n, sizeof(SortMbr), sort_lower_mbr);
    qsort(smu, n, sizeof(SortMbr), sort_upper_mbr);
    
    minover = MAXREAL;
    mindead = MAXREAL;
    // for all possible distributions of sml and snu
    for (k = 0; k < n - 2*m1 + 1; k++)
    {
        // lower sort
	// now calculate margin of R1
	// initialize mbr of R1 
        dead = 0.0;
	for (s = 0; s < 2*dimension; s += 2)
	{
	    rxmbr[s] =    MAXREAL;
	    rxmbr[s+1] = -MAXREAL;
	}
	for (l = 0; l < m1+k; l++)
	{
	    // calculate mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =   min(rxmbr[s],   sml[l].mbr[s]);
		rxmbr[s+1] = max(rxmbr[s+1], sml[l].mbr[s+1]);
	    }
            dead -= area(dimension, sml[l].mbr);
	}
        dead += area(dimension, rxmbr);
	
	// now calculate margin of R2
	// initialize mbr of R2 
	for (s = 0; s < 2*dimension; s += 2)
	{
	    rymbr[s] =    MAXREAL;
       	    rymbr[s+1] = -MAXREAL;
	}
	for ( ; l < n; l++)
	{
	    // calculate mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rymbr[s] =   min(rymbr[s],   sml[l].mbr[s]);
		rymbr[s+1] = max(rymbr[s+1], sml[l].mbr[s+1]);
	    }
            dead -= area(dimension, sml[l].mbr);
	}
        dead += area(dimension, rymbr);
	over = overlap(dimension, rxmbr, rymbr); 

        if ((over < minover) ||
	    (over == minover) && dead < mindead)
        {
            minover = over;
            mindead = dead;
            dist = m1+k;
            lu = TRUE;
        }

        // upper sort
	// now calculate margin of R1
	// initialize mbr of R1 
        dead = 0.0;
	for (s = 0; s < 2*dimension; s += 2)
	{
	    rxmbr[s] =    MAXREAL;
	    rxmbr[s+1] = -MAXREAL;
	}
	for (l = 0; l < m1+k; l++)
	{
	    // calculate mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =   min(rxmbr[s],   smu[l].mbr[s]);
		rxmbr[s+1] = max(rxmbr[s+1], smu[l].mbr[s+1]);
	    }
            dead -= area(dimension, smu[l].mbr);
	}
        dead += area(dimension, rxmbr);
	
	// now calculate margin of R2
	// initialize mbr of R2 
	for (s = 0; s < 2*dimension; s += 2)
	{
	    rymbr[s] =    MAXREAL;
	    rymbr[s+1] = -MAXREAL;
	}
	for ( ; l < n; l++)
	{
	    // calculate mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rymbr[s] =   min(rymbr[s],   smu[l].mbr[s]);
		rymbr[s+1] = max(rymbr[s+1], smu[l].mbr[s+1]);
	    }
            dead -= area(dimension, smu[l].mbr);
	}
        dead += area(dimension, rxmbr);
	over = overlap(dimension, rxmbr, rymbr); 

        if ((over < minover) ||
	    (over == minover) && dead < mindead)
        {
            minover = over;
            mindead = dead;
            dist = m1+k;
            lu = FALSE;
        }
    }

    // calculate best distribution
    *distribution = new int[n];    
    for (i = 0; i < n; i++)
    {
        if (lu)
            (*distribution)[i] = sml[i].index;
        else
            (*distribution)[i] = smu[i].index;
    }

    delete [] sml;
    delete [] smu;
    delete [] rxmbr;
    delete [] rymbr;


    return dist;
}


template <class DATA> 
int XTNode<DATA>::overlap_free_split(float **mbr, int *split_history, int **distribution, int *dim)
// teilt ein Array von mbrs in zwei Haelften und liefert in distribution
// zurueck, welche *m Eintraege in die neue Seite verschoben werden muessen
// hierfuer wird der X-Baum Split Algorithmus benutzt
{
#ifdef SHOWMBR
	split_000++;
#endif

    bool lu;
    int i, j, k, l, s, n, m1, dist, split_axis;
    int split_vector, num_axis, a;
    int axis[dimension];
    SortMbr *sml, *smu;
    float minmarg, marg, minover, mindead, dead, over, 
	*rxmbr, *rymbr;

    // how many nodes are used?
    n = get_num();

    // check split-history
    split_vector = split_history[0];
    for (i = 1; i < n; i++)
        split_vector = split_vector & split_history[i];

    if (split_vector == 0)
       return 0;


    // nodes have to be filled with at least one entry
    m1 = 1; 

    // sort by lower value of their rectangles
    // Indexarray aufbauen und initialisieren
    sml = new SortMbr[n];    
    smu = new SortMbr[n];    
    rxmbr = new float[2*dimension];
    rymbr = new float[2*dimension];

    
    num_axis = 0;
    for (i = 0; i < dimension; i++)
    {
        if ((split_vector & 1) == 1) {
            axis[num_axis] = i;
            num_axis++;
        }
        split_vector = split_vector >> 1;
    }

    // choose split axis from the above pre-selection
    minmarg = MAXREAL;
    for (a = 0; a < num_axis; a++)
    // for each axis
    {
        i = axis[a];

        for (j = 0; j < n; j++)
        {
            sml[j].index = smu[j].index = j;
            sml[j].dimension = smu[j].dimension = i;
            sml[j].mbr = smu[j].mbr = mbr[j];
        }

        // Sort by lower and upper value perpendicular axis_i
      	qsort(sml, n, sizeof(SortMbr), sort_lower_mbr);
        qsort(smu, n, sizeof(SortMbr), sort_upper_mbr);

        marg = 0.0;
        // for all possible distributions of sml
        for (k = 0; k < n - 2*m1 + 1; k++)
        {
            // now calculate margin of R1
	    // initialize mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =    MAXREAL;
		rxmbr[s+1] = -MAXREAL;
	    }
            for (l = 0; l < m1+k; l++)
            {
                // calculate mbr of R1 
		for (s = 0; s < 2*dimension; s += 2)
	        {
	            rxmbr[s] =   min(rxmbr[s],   sml[l].mbr[s]);
	            rxmbr[s+1] = max(rxmbr[s+1], sml[l].mbr[s+1]);
                }
            }
	    marg += margin(dimension, rxmbr); 

            // now calculate margin of R2
	    // initialize mbr of R2 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =    MAXREAL;
		rxmbr[s+1] = -MAXREAL;
	    }
            for ( ; l < n; l++)
            {
                // calculate mbr of R1 
		for (s = 0; s < 2*dimension; s += 2)
	        {
	            rxmbr[s] =   min(rxmbr[s],   sml[l].mbr[s]);
	            rxmbr[s+1] = max(rxmbr[s+1], sml[l].mbr[s+1]);
                }
            }
	    marg += margin(dimension, rxmbr); 
        }

        // for all possible distributions of smu
       	for (k = 0; k < n - 2*m1 + 1; k++)
        {
            // now calculate margin of R1
	    // initialize mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =    MAXREAL;
		rxmbr[s+1] = -MAXREAL;
	    }
            for (l = 0; l < m1+k; l++)
            {
                // calculate mbr of R1 
		for (s = 0; s < 2*dimension; s += 2)
	        {
	            rxmbr[s] =   min(rxmbr[s],   smu[l].mbr[s]);
	            rxmbr[s+1] = max(rxmbr[s+1], smu[l].mbr[s+1]);
                }
            }
	    marg += margin(dimension, rxmbr); 

            // now calculate margin of R2
	    // initialize mbr of R2 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =    MAXREAL;
		rxmbr[s+1] = -MAXREAL;
	    }
            for ( ; l < n; l++)
            {
                // calculate mbr of R1 
		for (s = 0; s < 2*dimension; s += 2)
	        {
	            rxmbr[s] =   min(rxmbr[s],   smu[l].mbr[s]);
	            rxmbr[s+1] = max(rxmbr[s+1], smu[l].mbr[s+1]);
                }
            }
	    marg += margin(dimension, rxmbr); 
        }

        // actual margin better than optimum?
        if (marg < minmarg)
        {
            split_axis = i;
            minmarg = marg;
        }

    }

    /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
    *dim = split_axis;

    
    // choose best distribution for split axis
    for (j = 0; j < n; j++)
    {
	sml[j].index = smu[j].index = j;
	sml[j].dimension = smu[j].dimension = split_axis;
	sml[j].mbr = smu[j].mbr = mbr[j];
    }
	       
    // Sort by lower and upper value perpendicular split axis
    qsort(sml, n, sizeof(SortMbr), sort_lower_mbr);
    qsort(smu, n, sizeof(SortMbr), sort_upper_mbr);
    
    minover = MAXREAL;
    mindead = MAXREAL;
    // for all possible distributions of sml and snu
    for (k = 0; k < n - 2*m1 + 1; k++)
    {
        // lower sort
	// now calculate margin of R1
	// initialize mbr of R1 
        dead = 0.0;
	for (s = 0; s < 2*dimension; s += 2)
	{
	    rxmbr[s] =    MAXREAL;
	    rxmbr[s+1] = -MAXREAL;
	}
	for (l = 0; l < m1+k; l++)
	{
	    // calculate mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =   min(rxmbr[s],   sml[l].mbr[s]);
		rxmbr[s+1] = max(rxmbr[s+1], sml[l].mbr[s+1]);
	    }
            dead -= area(dimension, sml[l].mbr);
	}
        dead += area(dimension, rxmbr);
	
	// now calculate margin of R2
	// initialize mbr of R2 
	for (s = 0; s < 2*dimension; s += 2)
	{
	    rymbr[s] =    MAXREAL;
       	    rymbr[s+1] = -MAXREAL;
	}
	for ( ; l < n; l++)
	{
	    // calculate mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rymbr[s] =   min(rymbr[s],   sml[l].mbr[s]);
		rymbr[s+1] = max(rymbr[s+1], sml[l].mbr[s+1]);
	    }
            dead -= area(dimension, sml[l].mbr);
	}
        dead += area(dimension, rymbr);
	over = overlap(dimension, rxmbr, rymbr); 

        if ((over < minover) ||
	    (over == minover) && dead < mindead)
        {
            minover = over;
            mindead = dead;
            dist = m1+k;
            lu = TRUE;
        }

        // upper sort
	// now calculate margin of R1
	// initialize mbr of R1 
        dead = 0.0;
	for (s = 0; s < 2*dimension; s += 2)
	{
	    rxmbr[s] =    MAXREAL;
	    rxmbr[s+1] = -MAXREAL;
	}
	for (l = 0; l < m1+k; l++)
	{
	    // calculate mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rxmbr[s] =   min(rxmbr[s],   smu[l].mbr[s]);
		rxmbr[s+1] = max(rxmbr[s+1], smu[l].mbr[s+1]);
	    }
            dead -= area(dimension, smu[l].mbr);
	}
        dead += area(dimension, rxmbr);
	
	// now calculate margin of R2
	// initialize mbr of R2 
	for (s = 0; s < 2*dimension; s += 2)
	{
	    rymbr[s] =    MAXREAL;
	    rymbr[s+1] = -MAXREAL;
	}
	for ( ; l < n; l++)
	{
	    // calculate mbr of R1 
	    for (s = 0; s < 2*dimension; s += 2)
	    {
		rymbr[s] =   min(rymbr[s],   smu[l].mbr[s]);
		rymbr[s+1] = max(rymbr[s+1], smu[l].mbr[s+1]);
	    }
            dead -= area(dimension, smu[l].mbr);
	}
        dead += area(dimension, rxmbr);
	over = overlap(dimension, rxmbr, rymbr); 

        if ((over < minover) ||
	    (over == minover) && dead < mindead)
        {
            minover = over;
            mindead = dead;
            dist = m1+k;
            lu = FALSE;
        }
    }

    // calculate best distribution
    *distribution = new int[n];    
    for (i = 0; i < n; i++)
    {
        if (lu)
            (*distribution)[i] = sml[i].index;
        else
            (*distribution)[i] = smu[i].index;
    }

    delete [] sml;
    delete [] smu;
    delete [] rxmbr;
    delete [] rymbr;


    return dist;
}
    

//////////////////////////////////////////////////////////////////////////////
// XTDirNode
//////////////////////////////////////////////////////////////////////////////

template <class DATA> 
XTDirNode<DATA>::XTDirNode(int _nodesize, XTree<DATA> *rt)
    : XTNode<DATA>(rt)
// according block does not yet exist
// _nodesize is the size of the new node
{
    int header_size, i;
    DirEntry<DATA> * d;

    nodesize = _nodesize;
    // allocate one more intern blocks than necessary to avoid memory conflicts in
    // read_from_buffer and write_to_buffer
    intern_block = new int[nodesize+1];

    // mal kurz einen Dateneintrag erzeugen und schauen, wie gross der wird..
    d = new DirEntry<DATA>(dimension, rt);

    // von der Blocklaenge geht die Headergroesse ab
    header_size = sizeof(bool) + sizeof(char) + sizeof(int) + sizeof(int) + sizeof(int);
    block_capacity = (rt->file->get_blocklength() - header_size) / d->get_size();
    capacity = block_capacity * nodesize;
    delete d;

    // Eintraege erzeugen, das geht mit einem Trick, da C++ beim 
    // Initialisieren von Objektarrays nur Defaultkonstruktoren kapiert.
    // Daher wird ueber globale Variablen die Information uebergeben.
    XTDirNode__dimension = dimension;
    XTDirNode__my_tree = my_tree;
    entries = new DirEntry<DATA>[capacity];

    // neue Plattenbloecke an File anhaengen
    for (i = 0; i < nodesize; i++)
    {
	intern_block[i] = rt->file->append_block(rt->block_buffer);
    }
    block = intern_block[0];

    rt->num_of_inodes ++;

    // Plattenblock muss auf jeden Fall neu geschrieben werden
    dirty = TRUE;
}


template <class DATA> 
XTDirNode<DATA>::XTDirNode(XTree<DATA> *rt, int _block)
    : XTNode<DATA>(rt)
// zugehoeriger Plattenblock existiert schon
{
    int header_size;
    int l;
    DirEntry<DATA> * d;

    // mal kurz einen Dateneintrag erzeugen und schauen, wie gross der wird..
    d = new DirEntry<DATA>(dimension, rt);

    // von der Blocklaenge geht die Headergroesse ab
    header_size = sizeof(bool) + sizeof(char) + sizeof(int) + sizeof(int) + sizeof(int);
    block_capacity = (rt->file->get_blocklength() - header_size) / d->get_size();
    delete d;

    // first reading the header of the first block, to know how many blocks exist
    block = _block;    
    rt->file->read_block(rt->block_buffer, block);
    read_header(rt->block_buffer);

    capacity = nodesize * block_capacity;
    // allocate one more intern blocks than necessary to avoid memory conflicts in
    // read_from_buffer and write_to_buffer
    intern_block = new int[nodesize+1];
    intern_block[0] = block;
    
    // Eintraege erzeugen, das geht mit einem Trick, da C++ beim
    // Initialisieren von Objektarrays nur Defaultkonstruktoren kapiert.
    // Daher wird ueber globale Variablen die Information uebergeben.
    XTDirNode__dimension = dimension;
    XTDirNode__my_tree = my_tree;
    entries = new DirEntry<DATA>[capacity];

    // zugehoerigen Plattenblock holen und Daten einlesen
    // dies kann nicht in XTNode::XTNode(..) passieren, weil
    // beim Aufruf des Basisklassenkonstruktors XTDirNode noch
    // gar nicht konstruiert ist, also auch keine Daten aufnehmen kann
    for (l = 0; l < nodesize; l++)
    {
	rt->file->read_block(rt->block_buffer, intern_block[l]);
	read_from_buffer(rt->block_buffer, l);
    }    

    // Plattenblock muss vorerst nicht geschrieben werden
    dirty = FALSE;
}


template <class DATA> 
XTDirNode<DATA>::~XTDirNode()
{
    int l;

    if (dirty)
    {
	// writing data to the according physical blocks
	for (l = 0; l < nodesize; l++)
        {
	    write_to_buffer(my_tree->block_buffer, l);
	    my_tree->file->write_block(my_tree->block_buffer, intern_block[l]);
	}    
    }

    delete [] entries;
}


template <class DATA> 
void XTDirNode<DATA>::read_header(char *buffer)
{
    int j;

    // Lesen, ob Sohnknoten eine Datenseite ist
    memcpy(&son_is_data, buffer, sizeof(bool));
    j = sizeof(bool);

    // Level des Knotens lesen
    memcpy(&level, &buffer[j], sizeof(char));
    j += sizeof(char);

    // Anzahl der belegten Eintraege lesen
    memcpy(&num_entries, &buffer[j], sizeof(int));
    j += sizeof(int);

    // Groesse des DirNodes
    memcpy(&nodesize, &buffer[j], sizeof(int));
    j += sizeof(int);
}


template <class DATA> 
void XTDirNode<DATA>::read_from_buffer(char *buffer, int offset)
// reads the data from buffer
// offset specifies the position of the entries and the block numbers of supernodes
// in entries[] respectively intern_block[]
{
    int i, j, s;

    // Lesen, ob Sohnknoten eine Datenseite ist
    memcpy(&son_is_data, buffer, sizeof(bool));
    j = sizeof(bool);

    // Level des Knotens lesen
    memcpy(&level, &buffer[j], sizeof(char));
    j += sizeof(char);

    // Anzahl der belegten Eintraege lesen
    memcpy(&num_entries, &buffer[j], sizeof(int));
    j += sizeof(int);

    // Groesse des DirNodes
    memcpy(&nodesize, &buffer[j], sizeof(int));
    j += sizeof(int);

    // Zeiger auf naechsten Block
    memcpy(&intern_block[offset+1], &buffer[j], sizeof(int));
    j += sizeof(int);   

    s = entries[0].get_size();
    for (i = 0; i < block_capacity; i++)
    {
	entries[i + offset*block_capacity].read_from_buffer(&buffer[j]);
	entries[i + offset*block_capacity].son_is_data = son_is_data;
	j += s;
    }
}

template <class DATA> 
void XTDirNode<DATA>::write_to_buffer(char *buffer, int offset)
// writes the data to buffer
// offset has the same function as in read_from_buffer
{
    int i, j, s;

    // Schreiben, ob Sohnknoten eine Datenseite ist
    memcpy(buffer, &son_is_data, sizeof(bool));
    j = sizeof(bool);

    // Level des Knotens schreiben
    memcpy(&buffer[j], &level, sizeof(char));
    j += sizeof(char);

    // Anzahl der belegten Eintraege schreiben
    memcpy(&buffer[j], &num_entries, sizeof(int));
    j += sizeof(int);

    // Groesse des DirNodes
    memcpy(&buffer[j], &nodesize, sizeof(int));
    j += sizeof(int);

    // Zeiger auf naechsten Block (bei supernodes)
    memcpy(&buffer[j], &intern_block[offset+1], sizeof(int));
    j += sizeof(int);   

    s = entries[0].get_size();
    for (i = 0; i < block_capacity; i++)
    {
	entries[i + offset*block_capacity].write_to_buffer(&buffer[j]);
	j += s;
    }
}

template <class DATA> 
void XTDirNode<DATA>::print()
{
    int i, n;

    n = get_num();
    for (i = 0; i < n ; i++)
    { 
        printf("(%4.1lf, %4.1lf, %4.1lf, %4.1lf)\n", 
	       entries[i].bounces[0],
	       entries[i].bounces[1],
	       entries[i].bounces[2],
	       entries[i].bounces[3]);
    }
    printf("level %d\n", level);
}

template <class DATA> 
int XTDirNode<DATA>::get_num_of_data()
{
    int i, n, sum;

    n = get_num();
    sum = 0;
    for (i = 0; i < n ; i++)
        sum += entries[i].num_of_data;
    return sum;
}


template <class DATA> 
float * XTDirNode<DATA>::get_mbr()
// liefert mbr des Directoryknotens
{
    int i, j, n;
    float *mbr;

    mbr = new float[2*dimension];
    for (i = 0; i < 2*dimension; i ++ )

	mbr[i] = entries[0].bounces[i];

    n = get_num();
    for (j = 1; j < n; j++)
    {
	for (i = 0; i < 2*dimension; i += 2)
	{
	    mbr[i]   = min(mbr[i],   entries[j].bounces[i]);
	    mbr[i+1] = max(mbr[i+1], entries[j].bounces[i+1]);
        }
    }

    return mbr;
}


template <class DATA>
float* XTDirNode<DATA>::calc_mbr(DirEntry<DATA> *de, int num_entries)
// Calculates the Minimal Bounding Rectangle for a set of Directory Entries 
{
    int i, j;
    float *mbr;
    
    mbr = new float[2*dimension];
    for (i = 0; i < 2*dimension; i++)
        mbr[i] = de[0].bounces[i];
    
    for (j = 0; j < num_entries; j++)
    {
        for (i = 0; i < 2*dimension; i += 2)
	{
	    mbr[i]   = min(mbr[i],   de[j].bounces[i]);
	    mbr[i+1] = max(mbr[i+1], de[j].bounces[i+1]);
	}
    }
    
    return mbr;
}


template <class DATA> 
void XTDirNode<DATA>::enter(DirEntry<DATA> *de)
{
    // ist ein Einfuegen ueberhaupt moeglich?
    if (get_num() > (capacity-1))
        error("XTDirNode::enter: called, but node is full", TRUE);
  
    // Eintrag an erste freie Stelle kopieren
    entries[num_entries] = *de;

    // jetzt gibts einen mehr
    num_entries++;

    // Eintrag freigeben, aber keine Nachfolgeknoten loeschen !!!!
    de->son_ptr = NULL;
    delete de;
}


template <class DATA>
void XTDirNode<DATA>::adapt_node(int mult)
// adapt node to a multiple of EXT_SIZE
{
    int i, n;
    int size, upper_border;
    int* intern_help;
    DirEntry<DATA> *new_entries;
        
    n = get_num();

    // insure adapting to a multiple of EXT_SIZE
    if (nodesize == 1)
    {
	size = 0;
	upper_border = 4;
    }
    else
    {
	size = nodesize;
	upper_border = EXT_SIZE*mult;
    }
    
    // adapt intern_block
    // allocate one more intern blocks than necessary to avoid memory conflicts in
    // read_from_buffer and write_to_buffer
    intern_help = new int[size+EXT_SIZE*mult+1];
    for (i = 0; i < nodesize; i++)
	intern_help[i] = intern_block[i];
    delete [] intern_block;
    intern_block = intern_help;

    // append new blocks to file
    for (i = 0; i < upper_border; i++)
	intern_block[nodesize+i] = my_tree->file->append_block(my_tree->block_buffer);

    nodesize = size + EXT_SIZE * mult;
    capacity = nodesize * block_capacity;
    
    new_entries = new DirEntry<DATA>[capacity];

    memcpy(new_entries, entries, n*sizeof(DirEntry<DATA>));

    // delete old entries
    for (i = 0; i < n; i++)
    {
	entries[i].bounces = NULL;
	entries[i].son_ptr = NULL;
    }
    delete [] entries;

    entries = new_entries;
}


template <class DATA> 
bool XTDirNode<DATA>::split(XTDirNode<DATA> **sn, int *dimen)
// splittet den aktuellen Knoten so auf, dass m mbrs nach sn verschoben
// werden und gibt TRUE zurueck, wenn der Split ausgefuehrt wird bzw. FALSE,
// wenn ein Supernode erstellt wird.
// In dimen wird die Split-Dimension an den aufrufenden Knoten zurueckgegeben, um dort
// die Split-History zu aktualisieren.
{
    int i, *distribution, dist, dist2, n, dim;
    float **mbr_array;
    float *r1, *r2;
    float a1, a2, o, overlap_relative;
    int *split_hist;
    DirEntry<DATA> *new_entries1, *new_entries2;
    int size_of_new_node;

#ifdef SHOWMBR
	split_000++;
#endif

    // no split, if node is supernode

    dim = 0;
	
    // wieviele sind denn nun belegt?
    n = get_num();

    // mbr_array allokieren und belegen
    mbr_array = new floatptr[n];
    split_hist = new int[n];
    for (i = 0; i < n; i++)
    {
	mbr_array[i] = entries[i].bounces;
	split_hist[i] = entries[i].history;
    }
    
    // Verteilung berechnen
    dist = XTNode<DATA>::topological_split(mbr_array, &distribution, &dim);

    // neues Datenarray erzeugen
    // --> siehe Konstruktor
    XTDirNode__dimension = dimension;
    XTDirNode__my_tree = my_tree;
    new_entries1 = new DirEntry<DATA>[capacity];
    new_entries2 = new DirEntry<DATA>[capacity];
    
    for (i = 0; i < dist; i++)
    {
	new_entries1[i] = entries[distribution[i]];
    }
    for (i = dist; i < n; i++)
    {
	new_entries2[i-dist] = entries[distribution[i]];
    }
    
    // Calculate the overlap size of the two mbrs
    r1 = calc_mbr(new_entries1, dist);
    r2 = calc_mbr(new_entries2, n-dist);
    o = overlap(dimension, r1, r2);
    a1 = area(dimension, r1);
    a2 = area(dimension, r2);
    overlap_relative = o / (a1+a2-o);
    
    if (overlap_relative > MAX_OVERLAP)
    {
	// topological split fails -> try overlap minimal split

        dist2 = XTNode<DATA>::overlap_free_split(mbr_array, split_hist, &distribution, &dim);
	
	if (dist2 != 0)
	{
	    // split_axis found
	    dist = dist2;
	    
	    // delete mbr_array and split history
	    delete [] mbr_array;
	    delete [] split_hist;
	    
	    // test for unbalanced nodes
	    if (((dist < (block_capacity*MIN_FANOUT)) || ((n-dist) < (block_capacity*MIN_FANOUT))))
	    {
		// dirnode has to be adapted to the size of the next greater multiple of EXT_SIZE
		adapt_node (((n/block_capacity+1) / EXT_SIZE) + 1);
		
		for (i = 0; i < n; i++)
		{
		    new_entries1[i].bounces = NULL;
		    new_entries1[i].son_ptr = NULL;                     
		    new_entries2[i].bounces = NULL;
		    new_entries2[i].son_ptr = NULL;                     
		}
		delete [] new_entries1;
		delete [] new_entries2;

                // Testausgabe
		printf("DIRECTORY SPLIT : Creating Supernode.\n");
		printf("Alter und neuer Knoten => Entries : %d, capacity : %d, nodesize : %d\n", num_entries, capacity, nodesize);
		
		return FALSE;
	    }
	    else
	    {
		// nodes are balanced
		
		for (i = 0; i < dist; i++)
		{
		    new_entries1[i] = entries[distribution[i]];
		}
		for (i = dist; i < n; i++)
		{
		    new_entries2[i-dist] = entries[distribution[i]];
		}
		
		for (i = 0; i < n; i++)
		{
		    entries[i].son_ptr = NULL;
		}
		delete [] entries;
		
	        // decides, whether brother has to be a supernode or a regular directory node
		size_of_new_node = ((((n-dist)/block_capacity+1) / EXT_SIZE) + 1)*EXT_SIZE;
		if ((n-dist) < block_capacity)
		{
		    size_of_new_node = 1;
		}

		*sn = new XTDirNode<DATA>(size_of_new_node, my_tree);
	    
		(*sn)->son_is_data = son_is_data;
		(*sn)->level = level;

		// neue Datenarrays kopieren
		entries = new_entries1;
		(*sn)->entries = new_entries2;
		
		// Anzahl der Eintraege berichtigen
		num_entries = dist;
		(*sn)->num_entries = n - dist;  // muss wegen Rundung so bleiben !!

                // Testausgabe
                printf("DIRECTORY SPLIT : Balanced overlap free split.\n");
		printf("Alter Knoten => Entries : %d, capacity : %d, nodesize : %d\n", num_entries, capacity, nodesize);
		printf("Neuer Knoten => Entries : %d, capacity : %d, nodesize : %d\n", (*sn)->num_entries, (*sn)->capacity, (*sn)->nodesize);
                 
		*dimen = dim;
	    }                
	}
    }
    else
    {
	// Datenarrays freigeben
	// da die Nachfolgeknoten aber nicht geloescht werden sollen
	// werden vor dem Aufruf von delete noch alle Pointer auf NULL gesetzt
	for (i = 0; i < n; i++)
	{
	    entries[i].son_ptr = NULL;
	}
	delete [] entries;

	// decides, whether brother has to be a supernode or a regular directory node
	if ((n-dist) >= block_capacity)
	{
	    *sn = new XTDirNode<DATA>( (((((n-dist)/block_capacity+1) / EXT_SIZE) + 1)*EXT_SIZE), my_tree);
	}
	else
	{
	    *sn = new XTDirNode<DATA>( 1, my_tree);
	}
   
	(*sn)->son_is_data = son_is_data;
	(*sn)->level = level;
		
	// neue Datenarrays kopieren
	entries = new_entries1;
	(*sn)->entries = new_entries2;
	
	// Anzahl der Eintraege berichtigen
	num_entries = dist;
	(*sn)->num_entries = n - dist;  // muss wegen Rundung so bleiben !!

        // Testausgabe
	printf("DIRECTORY SPLIT : Topological Split.\n");
	printf("Alter Knoten => Entries : %d, capacity : %d, nodesize : %d\n", num_entries, capacity, nodesize);
	printf("Neuer Knoten => Entries : %d, capacity : %d, nodesize : %d\n", (*sn)->num_entries, (*sn)->capacity, (*sn)->nodesize);

	*dimen = dim;
    }

    return TRUE;
}


template <class DATA> 
int XTDirNode<DATA>::choose_subtree(float *mbr)
{
    int i, j, n, follow, minindex, *inside, inside_count, *over;
    float *bmbr, old_o, o, omin, a, amin, f, fmin;

    n = get_num();

    // faellt d in einen bestehenden Eintrag ?
    inside_count = 0;
    inside = new int[n];
    over = new int[n];
    for (i = 0; i < n; i++)
    {
	switch (entries[i].section(mbr))
	{
	case INSIDE:
	    inside[inside_count++] = i;
	    break;
	}
    }

    if (inside_count == 1)
    // Fall 1: Rechteck faellt genau in einen Eintrag
	follow = inside[0];
    else if (inside_count > 1)
    // Fall 2: Rechteck faellt in mehrere Eintrag --> nimm den
    // mit der geringsten Flaeche (volumen!!!)
    {
	fmin = MAXFLOAT;
	//printf("Punkt in %d von %d MBRs \n",inside_count,n);
	for (i = 0; i < inside_count; i++)
	{
	    f = area(dimension, entries[inside[i]].bounces);
	    if (f < fmin)
	    {
		minindex = i;
      		fmin = f;
       	    }
       	}
	follow = inside[minindex];
    }
    else
    // Fall 3: Rechteck faellt in keinen Eintrag -->
    // fuer Knoten, die auf interne Knoten zeigen:
    // nimm den Eintrag, der am geringsten vergroessert wird;
    // bei gleicher Vergroesserung:
    // nimm den Eintrag, der die geringste Flaeche hat
    //
    // fuer Knoten, die auf Datenknoten zeigen:
    // nimm den, der die geringste Ueberlappung verursacht
    // bei gleicher Ueberlappung:
    // nimm den Eintrag, der am geringsten vergroessert wird;
    // bei gleicher Vergroesserung:
    // nimm den Eintrag, der die geringste Flaeche hat
    {
       	if (son_is_data)
	{
            omin = MAXREAL;
	    fmin = MAXREAL;
	    amin = MAXREAL;
	    for (i = 0; i < n; i++)
	    {
		// berechne die mbr, wenn mbr in entries[i] eingefuegt wird
		enlarge(dimension, &bmbr, mbr, entries[i].bounces);

		// calculate area and area enlargement
		a = area(dimension, entries[i].bounces);
		f = area(dimension, bmbr) - a;

		// calculate overlap before enlarging entry_i
		old_o = o = 0.0;
		for (j = 0; j < n; j++)
		{
		    if (j != i)
		    {
			old_o += overlap(dimension,
					 entries[i].bounces,
					 entries[j].bounces);
			o += overlap(dimension,
				     bmbr,
				     entries[j].bounces);
		    }
	        }
	        o -= old_o;

	        // is this entry better than the former optimum ?
	        if ((o < omin) ||
		    (o == omin && f < fmin) ||
		    (o == omin && f == fmin && a < amin))
	        {
	       	    minindex = i;
		    omin = o;
		    fmin = f;
		    amin = a;
	        }
	        delete [] bmbr;
	    }
        }
        else
        {
	    fmin = MAXREAL;
	    amin = MAXREAL;
	    for (i = 0; i < n; i++)
	    {
	        // berechne die mbr, wenn mbr in entries[i] eingefuegt wird
	        enlarge(dimension, &bmbr, mbr, entries[i].bounces);

	        // calculate area and area enlargement
	        a = area(dimension, entries[i].bounces);
	        f = area(dimension, bmbr) - a;

	        // is this entry better than the former optimum ?
	        if ((f < fmin) ||
		    (f == fmin && a < amin))
	        {
	       	    minindex = i;
		    fmin = f;
	            amin = a;
	        }
	        delete [] bmbr;
	    }
        }

	// berechne die neue mbr und schreib sie in den Eintrag
	enlarge(dimension, &bmbr, mbr, entries[minindex].bounces);
	memcpy(entries[minindex].bounces, bmbr,
	       dimension * 2 * sizeof(float));
	follow = minindex;
	delete [] bmbr;

	// leider muss jetzt der Block auch geschrieben werden
	dirty = TRUE;
    }

    delete [] inside;
    delete [] over;

    return follow;
}


template <class DATA> 
R_OVERFLOW XTDirNode<DATA>::insert(DATA *d, XTNode<DATA> **sn, int *dim)
// dim references a variable in the father node, so the split history can be updated there
{
    int follow, dimen;
    XTNode<DATA> *succ, *new_succ;
    XTDirNode<DATA> *brother;
    DirEntry<DATA> *de;
    R_OVERFLOW ret;
    float *mbr,*nmbr;    

    dimen = 0;
    
    // Einfuegepfad waehlen
    mbr = d->get_mbr();
    follow = choose_subtree(mbr);
    delete [] mbr;

    // Sohn laden
    succ = entries[follow].get_son();

    // insert d into son
    ret = succ->insert(d, &new_succ, &dimen);
    *dim = dimen;
    if (ret != NONE)
    // if anything happend --> update bounces of entry "follow"
    {
        mbr = succ->get_mbr();
        memcpy(entries[follow].bounces, mbr, sizeof(float) * 2*dimension);
        delete [] mbr;
    }

    // recalculate # of succeeders in the tree
    entries[follow].num_of_data = succ->get_num_of_data();

    if (ret == SPLIT)
    // Einfuegen von d in Sohn hat zum Split des Sohns gefuehrt -->
    // neuen Sohn *new_succ einfuegen
    {
        if (get_num() == capacity)
     	    error("XTDataNode::insert: maximum capacity violation", TRUE);

        // update split-history of old son
	entries[follow].history = entries[follow].history | (1 << dimen);
	
	// create and insert new entry
        de = new DirEntry<DATA>(dimension, my_tree);
	nmbr = new_succ->get_mbr();
        memcpy(de->bounces, nmbr, 2*dimension*sizeof(float));
	delete [] nmbr;
        de->son = new_succ->block;
        de->son_ptr = new_succ;
        de->son_is_data = son_is_data;
        de->num_of_data = new_succ->get_num_of_data();
        
	// set split-history of new son
        de->history = entries[follow].history;
        enter(de);

        if (get_num() == (capacity - 1))
        // Knoten laeuft ueber --> Split
        // this happens already if the node is nearly filled
        // for the algorithms are more easy then
        {
	    if (split(&brother, &dimen) == TRUE)
	    {
		*dim = dimen;
		*sn = brother;
		printf("splitting node %d, creating %d\n", block, (*sn)->block);

		ret = SPLIT;
	    }
	    else
	    {
               // no split occured in son, but supernode was created
	       delete sn;
	       ret = SUPERNODE;
	    }
	}
        else
      	    ret = NONE;
    }
    // must write page
    dirty = TRUE;

    return ret;
}


template <class DATA> 
DATA * XTDirNode<DATA>::get(int i)
{
    int j, n, sum;
    XTNode<DATA> *sn;

    n = get_num();
    sum = 0;
    for (j = 0; j < n; j++)
    {
        sum += entries[j].num_of_data;

        if (sum > i)
	// i-th object is behind this node --> follow son
        {
            sn = entries[j].get_son();
            return sn->get(i - (sum - entries[j].num_of_data));
        }
    }

    return NULL;
}


template <class DATA> 
void XTDirNode<DATA>::region(float *mbr)
{
    int i, n;
    SECTION s;
    XTNode<DATA> *succ;

    n = get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        s = entries[i].section(mbr);
        if (s == INSIDE || s == OVERLAP)
        {
            // Rechteck ist interessant --> rekursiv weiter
            succ = entries[i].get_son();
            succ->region(mbr);
        }
    }
}


template <class DATA> 
void XTDirNode<DATA>::point_query(float *p)
{
    int i, n;
    XTNode<DATA> *succ;

    n = get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
	if (entries[i].is_inside(p))
        {
            // Rechteck ist interessant --> rekursiv weiter
            succ = entries[i].get_son();
            succ->point_query(p);
        }
    }
}


template <class DATA> 
void XTDirNode<DATA>::range_query(float *mbr, SortedLinList<DATA> *res)
{
    int i, n;
    SECTION s;
    XTNode<DATA> *succ;

#ifdef ZAEHLER
    page_access += my_tree->node_weight[level];
#endif

    n = get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        s = entries[i].section(mbr);
        if (s == INSIDE || s == OVERLAP)
        {
            // Rechteck ist interessant --> rekursiv weiter
            succ = entries[i].get_son();
            succ->range_query(mbr,res);
        }
    }
}


template <class DATA> 
void XTDirNode<DATA>::range_query(DATA *center, float radius, 
		SortedLinList<DATA> *res)
{
    int i, n;
    bool s;
    XTNode<DATA> *succ;

    #ifdef ZAEHLER
	page_access += my_tree->node_weight[level];
    #endif

    n = get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        s = entries[i].section_circle(center,radius);
        if (s)
        {
            // Rechteck ist interessant --> rekursiv weiter
            succ = entries[i].get_son();
            succ->range_query(center,radius,res);
        }
    }
}



#ifdef S3
template <class DATA> 
void XTDirNode<DATA>::neighbours(LinList<DATA> *sl,
                                 float eps,
                                 Result *rs,
                                 norm_ptr norm)
{
    int i, j;
    DATA *s;
    float *mbr;
    XTNode<DATA> *succ;

    mbr = new float[2*dimension];

    for (i = 0; i < get_num(); i++)
    {
        for (s = sl->get_first(); s != NULL; s = sl->get_next())
        {
           for (j = 0; j < dimension; j++)
           {
               mbr[2*j] = s->data[j] - eps;
               mbr[2*j+1] = s->data[j] + eps;
           }

           if (entries[i].section(mbr) != S_NONE)
           {
               // Rechteck ist interessant --> rekursiv weiter
               succ = entries[i].get_son();
               succ->neighbours(sl, eps, rs, norm);
           }
        }
    }

    delete [] mbr;
}
#endif // S3




template <class DATA>
void XTDirNode<DATA>::nearest_neighbour_search(DATA *QueryPoint, DATA *Nearest,
				float *nearest_distanz)
{
    float *minmax_distanz;		// Array fuer MINMAXDIST aller Eintraege
    int *indexliste;		// Liste (for Sorting and Prunching)
    int i,j,k,last,n;
    float akt_min_dist;		// minimal distanz computed upto now 
    float minmaxdist,mindist;
    
    BranchList *activebranchList;
    
#ifdef ZAEHLER
    page_access += my_tree->node_weight[level];
#endif
    
    n = get_num();
    
    activebranchList = new BranchList [n]; // Array erzeugen mit n Elementen
    
    for( i = 0; i < n; i++)
    {
	activebranchList[i].entry_number = i;
	activebranchList[i].minmaxdist = MINMAXDIST(QueryPoint,entries[i].bounces);
	activebranchList[i].mindist = MINDIST(QueryPoint,entries[i].bounces);	
    }	
    
    // sort branchList
    qsort(activebranchList,n,sizeof(BranchList),sort_min_dist); 
    
    // prune BrunchList
    last = prune_branch_list(nearest_distanz,activebranchList,n);
    
    for( i = 0; i < last; i++)
    {
	entries[activebranchList[i].entry_number].get_son()->nearest_neighbour_search(QueryPoint, Nearest, nearest_distanz);
 	
	last = prune_branch_list(nearest_distanz,activebranchList,last);
    }
    
    delete [] activebranchList;
}

template <class DATA>
void XTDirNode<DATA>::nearest_neighbour_search(DATA *QueryPoint, 
				SortedLinList<DATA> *res,
				float *nearest_distanz)
{
    float *minmax_distanz;		// Array fuer MINMAXDIST aller Eintraege
    int *indexliste;		// Liste (for Sorting and Prunching)
    int i,j,k,last,n;
    float akt_min_dist;		// minimal distanz computed upto now 
    float minmaxdist,mindist;
    
    BranchList *activebranchList;
    
#ifdef ZAEHLER
    page_access += my_tree->node_weight[level];
#endif

    n = get_num();
    
    k = res->get_num(); 	// wird haben eine k-nearest-Narbor-Query
    
    *nearest_distanz = res->get(k-1)->distanz;  // der aktuell letzte 
                                                // naechste Nachbar wird
                                                // versucht zu ersetzen.

    activebranchList = new BranchList [n]; // Array erzeugen mit n Elementen
 
    for( i = 0; i < n; i++)
    {
	activebranchList[i].entry_number = i;
	activebranchList[i].minmaxdist = MINMAXDIST(QueryPoint,entries[i].bounces);
	activebranchList[i].mindist = MINDIST(QueryPoint,entries[i].bounces);	
    }	

    // sort_branch_list
    qsort(activebranchList,n,sizeof(BranchList),sort_min_dist); 
 
    // prune_branch_list
    last = prune_branch_list(nearest_distanz,activebranchList,n);

    for( i = 0; i < last; i++)
    {
	entries[activebranchList[i].entry_number].get_son()->nearest_neighbour_search(QueryPoint, res, nearest_distanz);
 	
	last = prune_branch_list(nearest_distanz,activebranchList,last);
    }

    delete [] activebranchList;
}


template <class DATA>
void XTDirNode<DATA>::overlapping(float *p, int *nodes_t)
{
    int i, n;
    XTNode<DATA> *succ;

    // ein Knoten mehr besucht
    nodes_t[level]++;

    n = get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        if (entries[i].is_inside(p))
        {
            // Rechteck ist interessant --> rekursiv weiter
            succ = entries[i].get_son();
            succ->overlapping(p, nodes_t);
        }
    }
}

template <class DATA>
void XTDirNode<DATA>::nodes(int *nodes_a)
{
    int i, n;
    XTNode<DATA> *succ;

    // ein Knoten mehr besucht
    nodes_a[level]++;

    n = get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
	succ = entries[i].get_son();
	succ->nodes(nodes_a);
    }
}

template <class DATA> 
void XTDirNode<DATA>::write_info(FILE *f)
{
    int i,j,n;

    float *mbr;

    mbr = get_mbr();
    fprintf(f,"%d\n",level+1);
    fprintf(f,"move %f %f\n",mbr[0],mbr[2]);
    fprintf(f,"draw %f %f\n",mbr[1],mbr[2]);
    fprintf(f,"draw %f %f\n",mbr[1],mbr[3]);
    fprintf(f,"draw %f %f\n",mbr[0],mbr[3]);
    fprintf(f,"draw %f %f\n",mbr[0],mbr[2]);
    fprintf(f,"\n");
    delete [] mbr;

    n = get_num();
    for( i = 0; i < n ; i++)
	entries[i].get_son()->write_info(f);

}

//////////////////////////////////////////////////////////////////////////////
// XTDataNode
//////////////////////////////////////////////////////////////////////////////

template <class DATA> 
XTDataNode<DATA>::XTDataNode(XTree<DATA> *rt)
    : XTNode<DATA>(rt)
// zugehoeriger Plattenblock muss erst erzeugt werden
{
    int header_size;
    DATA * d;

    level = 0;

    // mal kurz einen Dateneintrag erzeugen und schauen, wie gross der wird..
    d = new DATA(&dimension);

    // von der Blocklaenge geht die Headergroesse ab
    header_size = sizeof(char) + sizeof(int);
    capacity = (rt->file->get_blocklength() - header_size) / d->get_size();

    delete d;

    // Eintraege erzeugen; das geht mit einem Trick, da C++ beim 
    // Initialisieren von Objektarrays nur Defaultkonstruktoren kapiert.
    // Daher wird ueber statische Variablen die Information uebergeben.
    XTDataNode__dimension = dimension;
    data = new DATA[capacity];

    // neuen Plattenblock an File anhaengen
    block = rt->file->append_block(rt->block_buffer);

    rt->num_of_dnodes ++;

    // Plattenblock muss auf jeden Fall neu geschrieben werden
    dirty = TRUE;
}

template <class DATA> 
XTDataNode<DATA>::XTDataNode(XTree<DATA> *rt, int _block)
    : XTNode<DATA>(rt)
// zugehoeriger Plattenblock existiert schon
{
    int header_size;
    DATA *d;

    // mal kurz einen Dateneintrag erzeugen und schauen, wie gross der wird..
    d = new DATA(&dimension);

    // von der Blocklaenge geht die Headergroesse ab
    header_size = sizeof(char) + sizeof(int);
    capacity = (rt->file->get_blocklength() - header_size) / d->get_size();

    delete d;

    // Eintraege erzeugen; das geht mit einem Trick, da C++ beim 
    // Initialisieren von Objektarrays nur Defaultkonstruktoren kapiert.
    // Daher wird ueber statische Variablen die Information uebergeben.
    XTDataNode__dimension = dimension;
    data = new DATA[capacity];

    // zugehoerigen Plattenblock holen und Daten einlesen
    // dies kann nicht in XTNode::XTNode(..) passieren, weil
    // beim Aufruf des Basisklassenkonstruktors XTDirNode noch
    // gar nicht konstruiert ist, also auch keine Daten aufnehmen kann
    block = _block;
    rt->file->read_block(rt->block_buffer, block);
    read_from_buffer(rt->block_buffer);

    // Plattenblock muss vorerst nicht geschrieben werden
    dirty = FALSE;
}

template <class DATA> 
XTDataNode<DATA>::~XTDataNode()
{
    if (dirty)
    {
	// Daten auf zugehoerigen Plattenblock schreiben
	write_to_buffer(my_tree->block_buffer);
	my_tree->file->write_block(my_tree->block_buffer, block);
    }

    delete [] data;
}

template <class DATA> 
void XTDataNode<DATA>::read_from_buffer(char *buffer)
{
    int i, j, s;

    // Level des Knotens lesen
    memcpy(&level, buffer, sizeof(char));
    j = sizeof(char);

///////////////////////////////////////////////////////////
    // da das Abspeichern des levels voellig ueberfluessig ist
    // und level manchmal falsch in den Dateien steht, dieser HACK:
    level = 0;
///////////////////////////////////////////////////////////

    // Anzahl der belegten Eintraege lesen
    memcpy(&num_entries, &buffer[j], sizeof(int));
    j += sizeof(int);

    s = data[0].get_size();
    for (i = 0; i < num_entries; i++)
    {
	data[i].read_from_buffer(&buffer[j]);
	j += s;
    }
}

template <class DATA> 
void XTDataNode<DATA>::write_to_buffer(char *buffer)
{
    int i, j, s;

    // Level des Knotens schreiben
    level = 0;
    memcpy(&level, buffer, sizeof(char));
    j = sizeof(char);

    // Anzahl der belegten Eintraege schreiben
    memcpy(&buffer[j], &num_entries, sizeof(int));
    j += sizeof(int);

    s = data[0].get_size();
    for (i = 0; i < num_entries; i++)
    {
	data[i].write_to_buffer(&buffer[j]);
	j += s;
    }
}

template <class DATA> 
void XTDataNode<DATA>::print()
{
    int i, n;

    n = get_num();
    for (i = 0; i < n ; i++)
    { 
        printf("(%4.1lf, %4.1lf)\n", 
	       data[i].data[0],
	       data[i].data[1]);
    }
    printf("level %d\n", level);
}


template <class DATA> 
int XTDataNode<DATA>::get_num_of_data()
{
    return get_num();
}

template <class DATA> 
float * XTDataNode<DATA>::get_mbr()
// liefert mbr des Datenknotens
{
    int i, j, n;
    float *mbr, *tm;

    mbr = data[0].get_mbr();
    n = get_num();
    for (j = 1; j < n; j++)
    {
	tm = data[j].get_mbr();
	for (i = 0; i < 2*dimension; i += 2)
	{
		mbr[i]   = min(mbr[i],   tm[i]);
		mbr[i+1] = max(mbr[i+1], tm[i+1]);
       	}
       	delete [] tm;
    }

    return mbr;
}


template <class DATA> 
void XTDataNode<DATA>::split(XTDataNode<DATA> *sn, int *dimen)
// splittet den aktuellen Knoten so auf, dass m mbrs nach sn verschoben
// werden und gibt die Split-Dimension zurueck
{
    int i, *distribution, dist, n, dim;
    float **mbr_array;
    DATA *new_data1;

#ifdef SHOWMBR
	split_000++;
#endif
    dim = 0;

    // wieviele sind denn nun belegt?
    n = get_num();

    // mbr_array allokieren und belegen
    mbr_array = new floatptr[n];
    for (i = 0; i < n; i++)
    {
        mbr_array[i] = data[i].get_mbr();
    }
    
    // Verteilung berechnen 
    dist = XTNode<DATA>::topological_split(mbr_array, &distribution, &dim);

    // neues Datenarray erzeugen
    // --> siehe Konstruktor
    XTDataNode__dimension = dimension;
    new_data1 = new DATA[capacity];

    for (i = 0; i < dist; i++)
        new_data1[i] = data[distribution[i]];

    for (i = dist; i < n; i++)
        sn->data[i-dist] = data[distribution[i]];

    // Datenarrays freigeben
    delete [] data;

    // neue Datenarrays kopieren
    data = new_data1;

    // Anzahl der Eintraege berichtigen
    num_entries = dist;
    sn->num_entries = n - dist;  // muss wegen Rundung so bleiben !!

    // mbr_array freigeben
    for (i = 0; i < n; i++)
	delete [] mbr_array[i];

    delete [] mbr_array;

    *dimen = dim;
}


template <class DATA> 
R_OVERFLOW XTDataNode<DATA>::insert(DATA *d, XTNode<DATA> **sn, int *dim)
// liefert FALSE, wenn die Seite gesplittet werden muss
{
    int i, last_cand, dimen;
    float *mbr, *center;
    SortMbr *sm;
    DATA *nd, *new_data;

    dimen = 0;
    
    if (get_num() == capacity)
	error("XTDataNode::insert: maximum capacity violation", TRUE);

    data[get_num()] = *d;
    num_entries ++;

    // Plattenblock zum Schreiben markieren
    dirty = TRUE;

    if (get_num() == (capacity - 1))
    // Knoten laeuft ueber --> Split
    // dies passiert schon beim vorletzten Datum, da dadurch 
    // die Algorithmen einfacher werden
    {
	*sn = new XTDataNode<DATA>(my_tree);
	(*sn)->level = level;
	split((XTDataNode *) *sn, &dimen);
	*dim = dimen;
	printf("splitting node %d, creating %d\n", block, (*sn)->block);

	return SPLIT;
    }
    else
        return NONE;
}

template <class DATA> 
DATA * XTDataNode<DATA>::get(int i)
{
    DATA *d;

    if (i >= get_num())
    // if there is no i-th object --> NULL
        return NULL;

    d = new DATA(&dimension);
    *d = data[i];
    
    return d;
}

template <class DATA> 
void XTDataNode<DATA>::region(float *mbr)
{
    int i, n,j;
    float *dmbr;

    n = get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        dmbr = data[i].get_mbr();
        if (section(dimension, dmbr, mbr))
	{
/*		printf("( ");
		for( j = 0; j < dimension; j++)
			printf(" %f",dmbr[j]);
			        	
		printf(" ) \n");
*/	}	
        delete [] dmbr;
    }
}



template <class DATA> 
void XTDataNode<DATA>::point_query(float *p)
{
    int i, n,j;
    float *dmbr;

#ifdef ZAEHLER
    page_access++;
#endif
    
    n = get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        dmbr = data[i].get_mbr();
        if (section(dimension, p, dmbr))
	{
/*		printf("( ");
		for( j = 0; j < dimension; j++)
			printf(" %f",dmbr[j]);
			        	
		printf(" ) \n");
*/	}	
	delete [] dmbr;
    }
}


template <class DATA> 
void XTDataNode<DATA>::range_query(float *mbr, SortedLinList<DATA> *res)
{
    int i, n,j;
    float *dmbr;
    DATA *point, *center;

#ifdef ZAEHLER
    page_access += my_tree->node_weight[0];
#endif

    n = get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        dmbr = data[i].get_mbr();
        if (section(dimension, dmbr, mbr))
	{
		point = new DATA(&dimension);
		for( j = 0; j < dimension; j++)
			point->data[j] = dmbr[2*j];

		// Center des MBRs ausrechnen
		center = new DATA(&dimension);
		for(i = 0; i < dimension; i++)
		    center->data[i] = (mbr[2*i] + mbr[2*i+1]) / 2;
		point->distanz = OBJECT_DIST(point,center);
		res->insert(point);		        	
	}	
        delete [] dmbr;
    }
}


template <class DATA> 
void XTDataNode<DATA>::range_query(DATA *center, float radius,
			 SortedLinList<DATA> *res)
{
    int i, n,j,k;
    float *dmbr;
    DATA *point;

#ifdef ZAEHLER
    page_access += my_tree->node_weight[0];
#endif

    n = get_num();
    for (i = 0; i < n; i++)
    // teste alle Rechtecke auf Ueberschneidung
    {
        dmbr = data[i].get_mbr();
        if (section_c(dimension, dmbr, center,radius))
	{
		point = new DATA(&dimension);
		for( j = 0; j < dimension; j++)
			point->data[j] = dmbr[2*j];
		point->distanz = OBJECT_DIST(point,center);
		res->insert(point);		        	
	}	
        delete [] dmbr;
    }
}


#ifdef S3
template <class DATA> 
void XTDataNode<DATA>::neighbours(LinList<DATA> *sl,
                                  float eps,
                                  res *rs,
                                  norm_ptr norm)
{
    int i;
    float value;
    DATA *s;

    for (i = 0; i < get_num(); i++)
    {
        for (s = sl->get_first(); s != NULL; s = sl->get_next())
        {
	   value = norm(s->data, data[i].data, s->dimension);

	   // wenn f im eps-Bereich um s liegt --> rekursiv weiter
           if (value < eps)
	       rs->insert_region(data[i], *s, value);
        }
    }
}
#endif // S3

template <class DATA>
void XTDataNode<DATA>::nearest_neighbour_search(DATA *QueryPoint, DATA *Nearest,
			float *nearest_distanz)
{
    int i,j;
    float nearest_dist,distanz;
    
    //nearest_dist = OBJECT_DIST(QueryPoint,Nearest);
    
#ifdef ZAEHLER
    page_access += my_tree->node_weight[0];
#endif

    for(i = 0; i < get_num(); i++)
    {
	distanz = OBJECT_DIST(QueryPoint,get(i));
	
	if (distanz <= *nearest_distanz)
	{
	    *nearest_distanz = distanz;
	    Nearest->distanz = distanz;  
	    for(j = 0; j < dimension; j++)
		Nearest->data[j] = get(i)->data[j];
	}
    }   	
}



template <class DATA>
void XTDataNode<DATA>::nearest_neighbour_search(DATA *QueryPoint, 
			SortedLinList<DATA> *res,
			float *nearest_distanz)
{
    int i,j,k;
    float nearest_dist,distanz;
    bool t;
    DATA *neu,*dummy,*element;
    
#ifdef ZAEHLER
    page_access += my_tree->node_weight[0];
#endif
    
    k = res->get_num();
    
    for (i = 0; i < get_num(); i++)
    {
	element = get(i);
	distanz = OBJECT_DIST(QueryPoint,element);
	
	if (distanz <= *nearest_distanz)
	{
	    // loesche letztes Elemente der res-Liste		
	    dummy = res->get(k-1);
	    t = res->erase();
	    
	    // Erzeuge neuen Punkt in der res-Liste
	    element->distanz = distanz;
	    res->insert(element);
	    
	    *nearest_distanz = res->get(k-1)->distanz;			
	}
	else
	  delete element;   	
    }
}


template <class DATA> 
void XTDataNode<DATA>::write_info(FILE *f)
{
    float *mbr;
  
    mbr = get_mbr();
    fprintf(f,"%d\n",level+1);
    fprintf(f,"move %f %f\n",mbr[0],mbr[2]);
    fprintf(f,"draw %f %f\n",mbr[1],mbr[2]);
    fprintf(f,"draw %f %f\n",mbr[1],mbr[3]);
    fprintf(f,"draw %f %f\n",mbr[0],mbr[3]);
    fprintf(f,"draw %f %f\n",mbr[0],mbr[2]);
    fprintf(f,"\n");
    delete [] mbr;
}

//////////////////////////////////////////////////////////////////////////////
// XTree
//////////////////////////////////////////////////////////////////////////////

template <class DATA> 
XTree<DATA>::XTree(char *fname, int _b_length, int cache_size, int _dimension)
// neuen R-Baum konstruieren
{
    file = new CachedBlockFile(fname, _b_length, cache_size);

    // header und block_buffer allokieren und header lesen
    header = new char [file->get_blocklength()];    
    block_buffer = new char [file->get_blocklength()];    

    // das ist zwar an sich hier Quatsch, aber damit wird die Variable
    // user_header richtig gesetzt, so dass beim Schreiben des Headers der
    // Aufrufer kein problem hat
    read_header(header);

    dimension = _dimension;
    root = 0;
    root_ptr = NULL;
    root_is_data = TRUE;
    num_of_data = num_of_inodes = num_of_dnodes = 0;

    root_ptr = new XTDataNode<DATA>(this);
    root = root_ptr->block;
}

template <class DATA> 
XTree<DATA>::XTree(char *fname, int cache_size)
// bereits vorhandenen R-Baum laden
{
    int j;

    // file existiert schon --> Blockgroesse wird im Konstruktor ignoriert
    file = new CachedBlockFile(fname, 0, cache_size);

    // header und block_buffer allokieren und header lesen
    header = new char [file->get_blocklength()];    
    block_buffer = new char [file->get_blocklength()];    
    file->read_header(header);

    read_header(header);

    root_ptr = NULL;

    if (get_num() == 0)
    // Baum war leer -> Datenknoten anlegen und d einfuegen 
    {
	root_ptr = new XTDataNode<DATA>(this);
	root = root_ptr->block;
	root_ptr->level = 0;
    }
    else 
	load_root();
}

template <class DATA> 
XTree<DATA>::~XTree()
{
    int j;

    write_header(header);

    file->set_header(header);
    delete [] header;
    delete [] block_buffer;

    if (root_ptr != NULL)
    {
        delete root_ptr;
        root_ptr = NULL;
    }
    delete file;

    printf("saved X-Tree containing %d internal, %d data nodes and %d data\n",
	   num_of_inodes, num_of_dnodes, num_of_data);
}

template <class DATA>
void XTree<DATA>::print_info(FILE *f)
{
    fprintf(f,"\n internal: %d data-nodes : %d Datas: %d \n", 
	num_of_inodes, num_of_dnodes, num_of_data);
}

template <class DATA>
void XTree<DATA>::read_header(char *buffer)
{
    int i;

    memcpy(&dimension, buffer, sizeof(dimension));
    i = sizeof(dimension);

    memcpy(&num_of_data, &buffer[i], sizeof(num_of_data));
    i += sizeof(num_of_data);

    memcpy(&num_of_dnodes, &buffer[i], sizeof(num_of_dnodes));
    i += sizeof(num_of_dnodes);

    memcpy(&num_of_inodes, &buffer[i], sizeof(num_of_inodes));
    i += sizeof(num_of_inodes);

    memcpy(&root_is_data, &buffer[i], sizeof(root_is_data));
    i += sizeof(root_is_data);

    memcpy(&root, &buffer[i], sizeof(root));
    i += sizeof(root);

    user_header = &buffer[i];
}


template <class DATA>
void XTree<DATA>::write_header(char *buffer)
{
    int i;

    memcpy(buffer, &dimension, sizeof(dimension));
    i = sizeof(dimension);

    memcpy(&buffer[i], &num_of_data, sizeof(num_of_data));
    i += sizeof(num_of_data);

    memcpy(&buffer[i], &num_of_dnodes, sizeof(num_of_dnodes));
    i += sizeof(num_of_dnodes);

    memcpy(&buffer[i], &num_of_inodes, sizeof(num_of_inodes));
    i += sizeof(num_of_inodes);

    memcpy(&buffer[i], &root_is_data, sizeof(root_is_data));
    i += sizeof(root_is_data);

    memcpy(&buffer[i], &root, sizeof(root));
    i += sizeof(root);

    user_header = &buffer[i];
}


template <class DATA> 
void XTree<DATA>::load_root()
{
    if (root_ptr == NULL)
    // Baum ist nicht leer und root nicht geladen -> root laden 
    {
	if (root_is_data)
	    root_ptr = new XTDataNode<DATA>(this, root);
        else
	    root_ptr = new XTDirNode<DATA>(this, root);
    }    
}

template <class DATA> 
void XTree<DATA>::insert(DATA* d)
{
    int i, j, dim, dim_bits;
    XTNode<DATA> *sn;
    XTDirNode<DATA> *nroot_ptr;
    int nroot;
    DirEntry<DATA> *de;
    R_OVERFLOW split_root;
    DATA *d_cand, *dc;
    float *nmbr;

    dim = 0;
    dim_bits = 0;
    
    // load root into memory
    load_root();

    split_root = root_ptr->insert(d, &sn, &dim);
    
    if (split_root == SPLIT)
    // insert has lead to split --> new root-page having two sons
    // root and sn
    {
	nroot_ptr = new XTDirNode<DATA>(1, this);
	nroot_ptr->son_is_data = root_is_data;
	nroot_ptr->level = root_ptr->level + 1;
	nroot = nroot_ptr->block;
	
	// mbr der alten root page holen und in neue einfuegen
	de = new DirEntry<DATA>(dimension, this);
	nmbr = root_ptr->get_mbr();
	memcpy(de->bounces, nmbr, 2*dimension*sizeof(float));
	delete [] nmbr;
	de->son = root_ptr->block;
	de->son_ptr = root_ptr;
	de->son_is_data = root_is_data;
	de->num_of_data = root_ptr->get_num_of_data();
	de->history = de->history | (1 << dim);
	dim_bits = de->history;
	nroot_ptr->enter(de);
	    
	// mbr der gesplitteten page holen und in neue einfuegen
	de = new DirEntry<DATA>(dimension, this);
	nmbr = sn->get_mbr();
	memcpy(de->bounces, nmbr, 2*dimension*sizeof(float));
	delete [] nmbr;
	de->son = sn->block;
	de->son_ptr = sn;
	de->son_is_data = root_is_data;
	de->num_of_data = sn->get_num_of_data();
	de->history = de->history | dim_bits;
	nroot_ptr->enter(de);
	    
	root = nroot;
	root_ptr = nroot_ptr;

	// ab dem 1. Split muss die root-page ein Directoryknoten sein
	root_is_data = FALSE;
    }

    // jetzt ist der Baum um eines groesser
    num_of_data++;

    // write nodes
/*
    delete root_ptr;
    root_ptr = NULL;
*/
}

template <class DATA> 
DATA * XTree<DATA>::get(int i)
{
    DATA * d;

    // load root into main memory
    load_root();

    d = root_ptr->get(i);

    return d;
}

template <class DATA> 
void XTree<DATA>::region(float *mbr)
{
    // load root node into main memory
    load_root();

    root_ptr->region(mbr);
}

template <class DATA> 
void XTree<DATA>::point_query(float *p)
{
    // load root node into main memory
    load_root();

    root_ptr->point_query(p);
}


#ifdef S3
template <class DATA> 
void XTree<DATA>::neighbours(LinList<DATA> *sl,
		    float eps,
		    Result *rs,
		    norm_ptr norm)
{
    // load root node into main memory
    load_root();

    root_ptr->neighbours(sl, eps, rs, norm);
}
#endif // S3


template <class DATA>
void XTree<DATA>::nearest_neighbour_search(DATA *QueryPoint, DATA *Nearest)
{
      float nearest_distanz;

      // load root node into main memory
      load_root();

      nearest_distanz = MAXREAL;

      root_ptr->nearest_neighbour_search(QueryPoint, Nearest, &nearest_distanz);
}


template <class DATA>
void XTree<DATA>::nearest_neighbour_search(DATA *QueryPoint,
		SortedLinList<DATA> *res)
{
      float nearest_distanz;

      // load root node into main memory
      load_root();

      nearest_distanz = MAXREAL;

      root_ptr->nearest_neighbour_search(QueryPoint,res,&nearest_distanz);
}

template <class DATA>
void XTree<DATA>::k_nearest_neighbour_search(DATA *QueryPoint, int k, 
		SortedLinList<DATA> *NeighborList)
{
    int i,l,n;
    float nearest_distanz;
    DATA *p;
    
    nearest_distanz = MAXREAL;
    
    load_root();
    
    // NeighborListListe vorbereiten 
    NeighborList->set_sorting(TRUE);	
    
    for(i = 0; i < k; i++)
    {
	p = new DATA(&dimension);
	for(l = 0; l < dimension; l++)
	    p->data[l] = MAXREAL;
	p->distanz = MAXREAL;	
	NeighborList->insert(p);
    }

    root_ptr->nearest_neighbour_search(QueryPoint,NeighborList,&nearest_distanz); 
}		

template <class DATA>
void XTree<DATA>::range_query(DATA *center, float radius,
			SortedLinList<DATA> *res)
{
    load_root();
    root_ptr->range_query(center,radius,res);
}


template <class DATA>
void XTree<DATA>::range_query(float *mbr,
			SortedLinList<DATA> *res)
{
    load_root();

    root_ptr->range_query(mbr,res);
}


template <class DATA>
void XTree<DATA>::overlapping(float *p, int *nodes_t)
{
    load_root();

    root_ptr->overlapping(p, nodes_t);
}


template <class DATA>
void XTree<DATA>::nodes(int *nodes_a)
{
    load_root();

    root_ptr->nodes(nodes_a);
}


template <class DATA>
void XTree<DATA>::write_info(FILE *f)
{
    load_root();
    root_ptr->write_info(f);
    fprintf(f,"\n");
}

template <class DATA>
void XTree<DATA>::set_node_weight(int cache_size)
{
    int nodes_tab[20];
    int i, j, rest;
    bool c;

    // calculate number of nodes per level
    for (i = 0; i < 20; i++)
	nodes_tab[i] = 0;

    nodes(nodes_tab);

    rest = cache_size;
    if (cache_size > 0)
	c = TRUE;

    for (i = root_ptr->level; i > 0; i--)
    {
	if (c)
	// cache is not fully filled 
	{
	    if (rest - nodes_tab[i] > 0)
	    // there is enough cache for this level
	    {
		rest -= nodes_tab[i];
		node_weight[i] = 0.0;
	    }
	    else
	    {
		// this is the last level to be chached and not all 
		// nodes of the level are in general in the cache
		node_weight[i] = 1.0 - ((float) rest / (float)nodes_tab[i]);
		rest = 0;
		c = FALSE;
	    }
	}
	else 
	// cache is full
	    node_weight[i] = 1.0;
    }

    // data nodes are not in cache
    node_weight[0] = 1.0;

    printf("cache configuration (%d blocks chache):\n", cache_size);
    for (i = 0; i < 10; i++)
	printf("level %d: %f of %d\n", i, node_weight[i], nodes_tab[i]);
}


#endif // __XTREE


